-------------------------------------------------------
--! @file
--! @author Julian Mendez <julian.mendez@cern.ch> (CERN - EP-ESE-BE)
--! @date April, 2017
--! @version 1.0
--! @brief IC control - Tx FIFO and encoding/serialization
-------------------------------------------------------

--! Include the IEEE VHDL standard library
library ieee;
--! Use STD_Logic to define vector types
use ieee.std_logic_1164.all;
--! Used to convert std_logic_vector to integer and manage fifo pointer
use ieee.numeric_std.all;

--! Include the work library
library work;
--! Use SCA Package to define specific types (vector arrays)
use work.SCA_PKG.all;

--! @brief IC_tx Entity - Tx FIFO and encoding/serialization
--! @details 
--! The IC_tx entity control the FIFO used to store the command to be sent
--! and perform the HDLC encoding in parallel of the serialization process.
entity ic_tx is
    generic (
        g_FIFO_DEPTH        : integer := 20                         --! Depth of the internal FIFO used to improve the timming performance
    );
   port (
        -- Clock and reset
        tx_clk_i            : in std_logic;                         --! Tx clock (Tx_frameclk_o from GBT-FPGA IP): must be a multiple of the LHC frequency
        tx_clk_en           : in std_logic;                         --! Tx clock enable signal must be used in case of multi-cycle path(tx_clk_i > LHC frequency). By default: always enabled
		  
        reset_i             : in std_logic;                         --! Reset all of the TX processes
        
        -- Status
        tx_ready_o          : out std_logic;                        --! IC core ready for a transaction
        
        -- Configuration
        GBTx_address_i      : in std_logic_vector(6 downto 0);      --! I2C address of the GBTx
        Register_addr_i     : in std_logic_vector(15 downto 0);     --! Address of the first register to be accessed
        nb_to_be_read_i     : in std_logic_vector(15 downto 0);     --! Number of words/bytes to be read (only for read transactions)
        
        -- Internal FIFO
        wr_i                : in std_logic;                         --! Request a write operation into the internal FIFO (Data to GBTx)
        data_i              : in std_logic_vector(7 downto 0);      --! Data to be written into the internal FIFO
        
        -- FSM Control
        start_read          : in std_logic;                         --! Request a write config. to the GBTx
        start_write         : in std_logic;                         --! Request a read config. to the GBTx

        -- IC lines
        tx_data_o           : out std_logic_vector(1 downto 0)      --! (TX) Array of bits to be mapped to the TX GBT-Frame (bits 83/84)
   );   
end ic_tx;

--! @brief IC_tx architecture - Tx FIFO and encoding/serialization
--! @details 
--! The IC_tx architecture describes the behavior of the two components used to manage the IC line
--! in order to send a specific command: write register/burst (consists of a specific header followed by the data) 
--! or read register/burst (consists of a specific header to inform the start register and the number of
--! byte to be read).
architecture behaviour of ic_tx is
    
    -- Constant
    constant DEL_PACKET     : std_logic_vector(7 downto 0)    := "01111110";
    
    -- Types
    --! The hdlcstate_t type defines the states available for the hdlcproc process
    type hdlcstate_t is (s0_waitForStart, 
                            s2_sendSOF, 
                            s3_sendCommand, 
                            s4_sendCRC, 
                            s5_sendEOF);
    
    -- Signals
    signal word_cnt         : integer range 0 to g_FIFO_DEPTH;          --! Counter used to count the number of data written into the FIFO (write command)
    signal word_cnt_v       : std_logic_vector(15 downto 0);            --! Same as word_cnt but with a std_logic_vector type
    signal wr_ptr           : integer range 7 to (7+g_FIFO_DEPTH);      --! Write pointer to manage the write position into the FIFO (reset value: 7, after the header)

    signal fifo             : reg8_arr(g_FIFO_DEPTH+7 downto 0);        --! FIFO memory: user defined depth + 7 bytes for the command header (GBTx documentation)
    
    signal cmd_len          : integer range 0 to (7+g_FIFO_DEPTH);      --! Save command length in register
    signal tx_parity        : std_logic_vector(7 downto 0);             --! Tx Data parity (xor between data words)
    signal parity           : std_logic_vector(7 downto 0);             --! Final parity register (xor between header words and tx_parity signal)
        
    
    signal hdlc_state       : hdlcstate_t := s0_waitForStart;           --! HDLC encoding FSM state register
    
    
    signal start_write_s0   : std_logic    := '0';                      --! Register to detect start_write pulse
    signal start_read_s0    : std_logic    := '0';                      --! Register to detect start_read pulse
    signal reset_fifo       : std_logic := '0';                         --! Reset synchronization signal between FIFO and HDLC FSM. FIFO is reseted when the command has been fully sent.
    

    signal offset_pos_sig         : integer range 0 to 8;
    signal reg_pos_sig            : integer range 0 to 8+g_FIFO_DEPTH;
    signal high_lvl_cnter_sig     : integer range 0 to 6;

begin                 --========####   Architecture Body   ####========-- 

    word_cnt_v <= std_logic_vector(to_unsigned((word_cnt), 16));
    
    -- TX Fifo
    --! The wr_fifo process generate, create the command and store the data
    --! to be sent. All of the signal are synchronous with the tx_clk_i
    --! clock and can be managed at full speed (no use of the clk_en signal)
    wr_fifo: process(tx_clk_i, reset_i)    
    begin
    
        if rising_edge(tx_clk_i) then

            if reset_i = '1' or reset_fifo = '1' then
                word_cnt    <= 0;
                wr_ptr      <= 7;
                cmd_len     <= 7;
                tx_parity   <= (others => '0');
            
            elsif start_write = '1' then
                fifo(0)     <= "00000000";
                fifo(1)     <= GBTx_address_i & '0';
                fifo(2)     <= "00000001";
                fifo(3)     <= (word_cnt_v(7 downto 0));
                fifo(4)     <= (word_cnt_v(15 downto 8));
                fifo(5)     <= (Register_addr_i(7 downto 0));
                fifo(6)     <= (Register_addr_i(15 downto 8));
            
                cmd_len     <= wr_ptr;
                
            elsif start_read = '1' then
                fifo(0)     <= "00000000";
                fifo(1)     <= GBTx_address_i & '1';
                fifo(2)     <= "00000001";
                fifo(3)     <= (nb_to_be_read_i(7 downto 0));
                fifo(4)     <= (nb_to_be_read_i(15 downto 8));
                fifo(5)     <= (Register_addr_i(7 downto 0));
                fifo(6)     <= (Register_addr_i(15 downto 8));
            
                cmd_len      <= 7;
                
            elsif wr_i = '1' then

                fifo(wr_ptr)        <= data_i;
                tx_parity           <= tx_parity xor data_i;
                word_cnt            <= word_cnt + 1;
                wr_ptr              <= wr_ptr + 1;

            end if;
			
        end if;

    end process;

    -- TX Serializer
    --! The hdlcproc process encodes and serializes the command into the two
    --! bits of the IC line.
    --! The process implements the following state machine:
    --!     * *s0_waitForStart*: Wait for a start_write or start_read instruction and calculate the parity byte
    --!     * *s2_sendSOF*: Send start of frame delimiter (serialization of the "01111110" byte)
    --!     * *s3_sendCommand*: Read the FIFO and encode/serialize the data
    --!     * *s4_sendCRC*: Send the parity byte
    --!     * *s5_sendEOF*: Send end of frame delimiter (serialization of the "01111110" byte)
    hdlcproc: process(tx_clk_i, reset_i)

        variable offset_pos         : integer range 0 to 8;
        variable reg_pos            : integer range 0 to 8+g_FIFO_DEPTH;
        variable high_lvl_cnter     : integer range 0 to 6;

    begin
    
        if rising_edge(tx_clk_i) then
        
			if reset_i = '1' then            
                offset_pos          := 0;
                high_lvl_cnter      := 0;
                
                reset_fifo          <= '1';
                tx_ready_o          <= '0';
                tx_data_o           <= "11";
                
                hdlc_state          <= s0_waitForStart;
                reg_pos             := 0;
            
			elsif tx_clk_en = '1' then
			
                reset_fifo <= '0';
                tx_ready_o <= '0';
                
                case hdlc_state is
                
                    when s0_waitForStart    =>    
                        
                        -- Send idle
                        reg_pos             := 0;
                        tx_ready_o          <= '1';

                        tx_data_o(1)        <= '1';
                        tx_data_o(0)        <= '1';

                        -- Check start signal
                        if start_write = '1' then

                            tx_ready_o <= '0';

                            -- Compute the parity word
                            parity <=
                                (   
                                    tx_parity                       xor 
                                    word_cnt_v(7 downto 0)          xor 
                                    word_cnt_v(15 downto 8)         xor 
                                    Register_addr_i(7 downto 0)     xor 
                                    Register_addr_i(15 downto 8)    xor 
                                    "00000001"
                                );
                            
                            tx_data_o(1)        <= DEL_PACKET(0);
                            tx_data_o(0)        <= DEL_PACKET(1);
                            offset_pos          := 2;
                            
                            hdlc_state <= s2_sendSOF;
                                                        
                        elsif start_read = '1' then
                                                        
                            tx_ready_o <= '0';
                            
                            -- Compute the parity word
                            parity <= 
                                (   
                                    nb_to_be_read_i(7 downto 0)     xor 
                                    nb_to_be_read_i(15 downto 8)    xor 
                                    Register_addr_i(7 downto 0)     xor 
                                    Register_addr_i(15 downto 8)    xor 
                                    "00000001"
                                );
                                
                            tx_data_o(1)        <= DEL_PACKET(0);
                            tx_data_o(0)        <= DEL_PACKET(1);
                            offset_pos          := 2;

                            hdlc_state <= s2_sendSOF;
                            
                        end if;

                    when s2_sendSOF         =>
                        -- Data have been copied into the vector buffer
                        --      -> The fifo can be reseted

                        -- Send delimiter without bit stuffing
                        tx_data_o(1)     <= DEL_PACKET(offset_pos);
                        tx_data_o(0)     <= DEL_PACKET(offset_pos+1);

                        offset_pos      := offset_pos + 2;

                        --make sure that the bit stuffing counter is always zeroed at the beginning of the transaction
                        high_lvl_cnter  := 0;

                        -- Delimiter has been fully sent?
                        if offset_pos > 7 then    
                            offset_pos  := 0;
                            hdlc_state  <= s3_sendCommand;
                        end if;

                    when s3_sendCommand     =>    
                                    
                    
                        -- Enough data to send 2 bits?
                        if offset_pos < 7 then
                        
                            -- (MSB) Bit stuffing?
                            if high_lvl_cnter >= 5 then
                            
                                tx_data_o(1)        <= '0';
                                high_lvl_cnter     := 0;
                            
                            -- (MSB) Send data
                            else
                            
                                tx_data_o(1)        <= fifo(reg_pos)(offset_pos);
                                if fifo(reg_pos)(offset_pos) = '1' then
                                    high_lvl_cnter := high_lvl_cnter + 1;
                                else
                                    high_lvl_cnter := 0;
                                end if;
                                
                                offset_pos         := offset_pos + 1;
                            
                            end if;
                        
                            -- (LSB) Bit stuffing?
                            if high_lvl_cnter >= 5 then
                            
                                tx_data_o(0)        <= '0';
                                high_lvl_cnter     := 0;
                            
                            -- (LSB) Send data
                            else
                            
                                tx_data_o(0)        <= fifo(reg_pos)(offset_pos);
                                if fifo(reg_pos)(offset_pos) = '1' then
                                    high_lvl_cnter := high_lvl_cnter + 1;
                                else
                                    high_lvl_cnter := 0;
                                end if;
                                
                                offset_pos         := offset_pos + 1;
                            
                            end if;
                            
                            -- End of command?
                            if offset_pos = 8 then
                            
                                offset_pos      := 0;
                                
                                if reg_pos < (cmd_len-1) then
                                    reg_pos         := reg_pos + 1;
                                else
                                    hdlc_state      <= s4_sendCRC;
                                end if;
                                
                            end if;
                        
                        -- 1 bit still have to been sent, then CRC
                        elsif offset_pos < 8 then
                        
                            -- (MSB) Bit stuffing?
                            if high_lvl_cnter >= 5 then
                            
                                tx_data_o(1)     <= '0';
                                tx_data_o(0)     <= fifo(reg_pos)(7);
                                if fifo(reg_pos)(offset_pos) = '1' then
                                    high_lvl_cnter     := high_lvl_cnter + 1;
                                else
                                    high_lvl_cnter     := 0;
                                end if;
                                
                                offset_pos      := 0;
                                
                                if reg_pos < (cmd_len-1) then
                                    reg_pos         := reg_pos + 1;
                                else
                                    hdlc_state      <= s4_sendCRC;
                                end if;
                                
                            -- (MSB) Send data
                            else
                            
                                tx_data_o(1)        <= fifo(reg_pos)(7);
                                if fifo(reg_pos)(offset_pos) = '1' then
                                    high_lvl_cnter := high_lvl_cnter + 1;
                                else
                                    high_lvl_cnter := 0;
                                end if;
                            
                                -- (LSB) Bit stuffing?
                                if high_lvl_cnter >= 5 then
                                    tx_data_o(0)     <= '0';
                                    high_lvl_cnter  := 0;

                                elsif reg_pos < (cmd_len-1) then
                                    
                                    reg_pos         := reg_pos + 1;
                                    
                                    tx_data_o(0)     <= fifo(reg_pos)(0);
                                    if fifo(reg_pos)(0) = '1' then
                                        high_lvl_cnter := high_lvl_cnter + 1;
                                    else
                                        high_lvl_cnter := 0;
                                    end if;
                                    
                                    offset_pos      := 1;
                                        
                                else
                                    tx_data_o(0)     <= parity(0);
                                    if parity(0) = '1' then
                                        high_lvl_cnter := high_lvl_cnter + 1;
                                    else
                                        high_lvl_cnter := 0;
                                    end if;

                                    offset_pos      := 1;    
                                    hdlc_state      <= s4_sendCRC;
                                
                                end if;
                                
                            end if;
                        
                        end if;
                    
                    when s4_sendCRC             =>
                    
                        -- Enough data to send 2 bits?
                        if offset_pos < 7 then
                        
                            -- (MSB) Bit stuffing?
                            if high_lvl_cnter >= 5 then
                            
                                tx_data_o(1)        <= '0';
                                high_lvl_cnter     := 0;
                            
                            -- (MSB) Send data
                            else
                            
                                tx_data_o(1)         <= parity(offset_pos);
                                if parity(offset_pos) = '1' then
                                    high_lvl_cnter := high_lvl_cnter + 1;
                                else
                                    high_lvl_cnter := 0;
                                end if;
                                
                                offset_pos          := offset_pos + 1;
                            
                            end if;
                        
                            -- (LSB) Bit stuffing?
                            if high_lvl_cnter >= 5 then
                            
                                tx_data_o(0)        <= '0';
                                high_lvl_cnter     := 0;
                            
                            -- (LSB) Send data
                            else
                            
                                tx_data_o(0)        <= parity(offset_pos);
                                if parity(offset_pos) = '1' then
                                    high_lvl_cnter := high_lvl_cnter + 1;
                                else
                                    high_lvl_cnter := 0;
                                end if;
                                
                                offset_pos      := offset_pos + 1;
                            
                            end if;
                            
                            -- End of command?
                            -- if the bit stuffing zero must be sent stay in this state
                            -- but with offset_pos=8
                            if offset_pos = 8 and high_lvl_cnter<5 then    
                                offset_pos      := 0;
                                reset_fifo      <= '1';
                                hdlc_state      <= s5_sendEOF;

                            end if;
                        
                        -- 1 bit still have to been sent, then CRC
                        elsif offset_pos < 8 then
                        
                            -- (MSB) Bit stuffing?
                            if high_lvl_cnter >= 5 then
                            
                                tx_data_o(1)     <= '0';
                                tx_data_o(0)     <= parity(offset_pos);
                                
                                if parity(offset_pos) = '1' then
                                    high_lvl_cnter     := 0;
                                end if;
                                
                                offset_pos      := 0;
                                reset_fifo      <= '1';
                                hdlc_state      <= s5_sendEOF;
                                
                            -- (MSB) Send data
                            else
                            
                                tx_data_o(1)        <= parity(offset_pos);
                                if parity(offset_pos) = '1' then
                                    high_lvl_cnter := high_lvl_cnter + 1;
                                else
                                    high_lvl_cnter := 0;
                                end if;
                            
                                -- (LSB) Bit stuffing?
                                if high_lvl_cnter >= 5 then
                                    tx_data_o(0)     <= '0';
                                    
                                    high_lvl_cnter  := 0;
                                    offset_pos      := 0;
                                    reset_fifo      <= '1';
                                    hdlc_state      <= s5_sendEOF;

                                else
                                    
                                    tx_data_o(0)     <= DEL_PACKET(0);
                                    offset_pos      := 1;
                                    reset_fifo      <= '1';
                                    hdlc_state      <= s5_sendEOF;
                                    
                                end if;
                            end if;

                        --the CRC is sent. This just have to add bit stuffing zero
                        else
                            tx_data_o(1)     <= '0';
                            tx_data_o(0)     <= DEL_PACKET(0);
                            high_lvl_cnter  := 0;
                            offset_pos      := 1;
                            reset_fifo      <= '1';
                            hdlc_state      <= s5_sendEOF;

                        end if;    
                        
                    when s5_sendEOF         =>
                        

                        -- (MSB) Send delimiter
                        tx_data_o(1)            <= DEL_PACKET(offset_pos);
                        offset_pos              := offset_pos + 1;
                        
                        -- Fully sent?
                        if offset_pos > 7 then
                        
                            tx_data_o(0)        <= '1';
                            hdlc_state          <= s0_waitForStart;
                                                            
                        else
                            
                            -- (LSB) Send delimiter
                            tx_data_o(0)        <= DEL_PACKET(offset_pos);
                            offset_pos          := offset_pos + 1;
                            
                            -- Fully sent?
                            if offset_pos > 7 then
                                hdlc_state         <= s0_waitForStart;
                            end if;
                            
                        end if;
                end case;

                offset_pos_sig  <= offset_pos;
                reg_pos_sig     <= reg_pos;
                high_lvl_cnter_sig <= high_lvl_cnter;
    
			end if;
			
        end if;
        
    end process;

end behaviour;
--============================================================================--
--############################################################################--
--============================================================================--