library ieee;
  use ieee.std_logic_1164.all;
  use ieee.numeric_std.all;               -- for type conversions


library general_cores;
  use general_cores.wishbone_pkg.all;

--! A wishbone bus controller driven by GBTx Internal Control (IC) lines
entity ic_wb_ctrl is
    generic (
        emu_id_g     : std_logic_vector(6 downto 0);
        send_wr_ack_g  : boolean := true;
        debug_g         : string := "false"
      );
    port (
      --! Wishbone CLK
      wb_clk_i       : in std_logic;
      --! Reset signal synchronous with Wishbone CLK
      wb_rst_i       : in std_logic;

      --! Wishbone output record
      wb_ms_o        : out t_wishbone_master_out;
      --! Wishbone input record
      wb_sm_i        : in t_wishbone_slave_out;

      --! Rx clock (Rx_frameclk_o from GBT-FPGA IP): must be a multiple of the LHC frequency
      rx_clk_i            : in  std_logic;                        
      --! Rx clock enable signal must be used in case of multi-cycle path(rx_clk_i > LHC frequency). By default: always enabled
      rx_clk_en           : in  std_logic;                        
    
      --! Reset RX datapath
      rx_reset_i          : in  std_logic;                                             
      --! (TX) Array of bits to be mapped to the TX GBT-Frame (bits 83/84)
      tx_data_o      : out std_logic_vector(1 downto 0);     

      --! a flag generated by GBT-FPGA it FEC receiver detected a transmission error
      --! it is uncertain if FEC was able to fix the error
      --! IC interface will drop uncertain frames for safety reasons
      rx_data_uncertain_i : in std_logic;
      
      --! (RX) Array of bits to be mapped to the RX GBT-Frame (bits 83/84)
      rx_data_i      : in  std_logic_vector(1 downto 0)
  );
end ic_wb_ctrl;

architecture ic_wb_ctrl_arc of ic_wb_ctrl is

  signal  cdc_rst_n         : std_logic;
  signal  rx_reset          : std_logic;

  signal  rep_fifo_wr       : std_logic;
  signal  rep_fifo_din      : std_logic_vector(7 downto 0);
  signal  rep_send          : std_logic;
  signal  rep_fifo_ready    : std_logic;

  signal  cmd_rd            : std_logic;
  signal  cmd_gbtid         : std_logic_vector(6 downto 0);
  signal  cmd_nbytes        : std_logic_vector(15 downto 0);
  signal  cmd_byteaddr      : std_logic_vector(15 downto 0);
  signal  cmd_fifo_empty    : std_logic;
  signal  cmd_fifo_rd       : std_logic;
  signal  cmd_fifo_dat      : std_logic_vector(7 downto 0);
  signal  cmd_parity_error  : std_logic;

  signal  wb_wr_r           : std_logic;
  signal  wb_cyc_r          : std_logic;
  signal  wb_addr_r         : unsigned(15 downto 0);
  signal  wb_rddat_r        : std_logic_vector(31 downto 0);
  signal  wb_wrdat_r        : std_logic_vector(31 downto 0);
  signal  byte_cnt_r        : unsigned(15 downto 0);
  signal  reg_byte_r        : unsigned(1 downto 0);


  constant c_protector_max     : natural := 31; 
  signal  protector_cnt        : natural range 0 to c_protector_max;
  signal  rst_from_protector   : std_logic;

                      --   0
  type state_et is (  idle_st, 
                   --         1           2                 3               4                   5              6
                      rdcmd_wbset_st, rdcmd_wbinc_st, rdcmd_w8wb_st, rdcmd_pushreply_st, rdcmd_send_st, rdcmd_wbincrd_st, 
                  --        7               8               9               10                11                12                13  
                      wrcmd_wbset_st, wrcmd_wbinc_st, wrcmd_w8rdwb_st, wrcmd_w8wrwb_st, wrcmd_wbmodify_st, wrcmd_wr_ack_st, wrcmd_wr_err_st, wrcmd_send_st);
  signal  crx_state_r, crx_state_r2      : state_et;

  --! Wishbone output record
  signal crx_wb_ms        : t_wishbone_master_out;
  --! Wishbone input record
  signal crx_wb_sm        : t_wishbone_slave_out;
  
  attribute MARK_DEBUG : string;
  attribute MARK_DEBUG of crx_state_r2  : signal is debug_g;
  attribute MARK_DEBUG of crx_wb_ms     : signal is debug_g;
  attribute MARK_DEBUG of crx_wb_sm     : signal is debug_g;
  
  attribute MARK_DEBUG of cmd_rd          : signal is debug_g;
  attribute MARK_DEBUG of cmd_nbytes      : signal is debug_g;
  attribute MARK_DEBUG of cmd_byteaddr    : signal is debug_g;
  attribute MARK_DEBUG of cmd_fifo_empty  : signal is debug_g;
  attribute MARK_DEBUG of cmd_fifo_rd     : signal is debug_g;
  attribute MARK_DEBUG of cmd_fifo_dat    : signal is debug_g;
  
  attribute MARK_DEBUG of rep_fifo_wr     : signal is debug_g;
  attribute MARK_DEBUG of rep_fifo_din    : signal is debug_g;
  attribute MARK_DEBUG of rep_send        : signal is debug_g;
  attribute MARK_DEBUG of rep_fifo_ready  : signal is debug_g;
  
  attribute MARK_DEBUG of byte_cnt_r        : signal is debug_g;
  attribute MARK_DEBUG of reg_byte_r        : signal is debug_g;
  attribute MARK_DEBUG of cmd_parity_error  : signal is debug_g;

begin

  crx_wb_ms.cyc   <= wb_cyc_r;
  crx_wb_ms.stb   <= wb_cyc_r;
  crx_wb_ms.adr   <= std_logic_vector(resize(wb_addr_r, 32));
  crx_wb_ms.sel   <= "1111";
  crx_wb_ms.we    <= wb_wr_r;
  crx_wb_ms.dat   <= wb_wrdat_r;

  rx_reset  <= rx_reset_i or rst_from_protector;
 
  fsm : process (rx_clk_i)

    variable regbyte_v    : integer;
    variable wb_dat_new_v : std_logic_vector(31 downto 0);

  begin
    if (rising_edge(rx_clk_i)) then
      if rx_reset='1' then 
        crx_state_r   <= idle_st;
        wb_wr_r                   <= '0';
        wb_cyc_r                  <= '0';
        rep_fifo_wr               <= '0';
        rep_send                  <= '0';
        cmd_fifo_rd               <= '0';

      else

        crx_state_r2               <= crx_state_r;
        wb_wr_r                   <= '0';
        wb_cyc_r                  <= '0';
        rep_fifo_wr               <= '0';
        rep_send                  <= '0';
        cmd_fifo_rd               <= '0';

        case(crx_state_r) is
          when idle_st => 
            if cmd_rd='1' then
              crx_state_r         <= rdcmd_wbset_st;

            elsif cmd_fifo_empty='0' then
                crx_state_r       <= wrcmd_wbset_st;

            end if;

          -----------------------------------------------------------------
          --------------------- IC Read -----------------------------------
          -----------------------------------------------------------------

          when rdcmd_wbset_st =>
            crx_state_r           <= rdcmd_w8wb_st;

            wb_addr_r         <= "00"&unsigned(cmd_byteaddr(15 downto 2));
            reg_byte_r        <= unsigned(cmd_byteaddr(1 downto 0));
            byte_cnt_r        <= (others => '0');
          
          when rdcmd_wbinc_st =>
            crx_state_r           <= rdcmd_w8wb_st;
            wb_addr_r             <= wb_addr_r+1;

          when rdcmd_w8wb_st =>
            
            if crx_wb_sm.ack='1' then
              crx_state_r         <= rdcmd_pushreply_st;
              wb_cyc_r            <= '0';

            -- don't answer to the read
            -- I can't send less bytes than master expects
            -- because he might expect 1 byte (I can't send 0 bytes)
            elsif crx_wb_sm.err='1' then
              crx_state_r         <= idle_st;
              wb_cyc_r            <= '0';

            else
              wb_cyc_r          <= '1';
              
            end if;

            wb_wr_r           <= '0';
            wb_rddat_r        <= crx_wb_sm.dat;

          when rdcmd_pushreply_st => 
            if byte_cnt_r<unsigned(cmd_nbytes)-1 then 
              if reg_byte_r(1 downto 0)="11" then
                crx_state_r       <= rdcmd_wbinc_st;

              else 
                crx_state_r       <= crx_state_r;

              end if;
            else
              crx_state_r         <= rdcmd_send_st;

            end if;


            reg_byte_r          <= reg_byte_r+1;
            regbyte_v           := to_integer(reg_byte_r);

            rep_fifo_din        <= wb_rddat_r(8*(regbyte_v+1)-1 downto 8*regbyte_v);
            rep_fifo_wr         <= '1';
            byte_cnt_r          <= byte_cnt_r+1;



          when rdcmd_send_st =>
            crx_state_r   <= idle_st;

            rep_send      <= '1';

          -----------------------------------------------------------------
          --------------------- IC Write -----------------------------------
          -----------------------------------------------------------------

          when wrcmd_wbset_st =>
            crx_state_r       <= wrcmd_w8rdwb_st;

            wb_addr_r         <= "00"&unsigned(cmd_byteaddr(15 downto 2));
            reg_byte_r        <= unsigned(cmd_byteaddr(1 downto 0));
            byte_cnt_r        <= (others => '0');
            cmd_fifo_rd       <= '1';
          
          when wrcmd_wbinc_st =>
            crx_state_r           <= wrcmd_w8rdwb_st;
            wb_addr_r             <= wb_addr_r+1;

          when wrcmd_w8rdwb_st =>
            

            if crx_wb_sm.ack='1' then
              crx_state_r         <= wrcmd_wbmodify_st;
              wb_cyc_r            <= '0';
            
            elsif crx_wb_sm.err='1' then
              wb_cyc_r            <= '0';
              if send_wr_ack_g then
                crx_state_r         <= wrcmd_wr_err_st;

              else
                crx_state_r         <= idle_st;
              
              end if;
            
            else
              wb_cyc_r          <= '1';

            end if;

            wb_wr_r           <= '0';
            wb_rddat_r        <= crx_wb_sm.dat;
          
          when wrcmd_w8wrwb_st =>
            
            if crx_wb_sm.ack='1' then

              wb_cyc_r            <= '0';
              wb_wr_r             <= '0';

              reg_byte_r          <= reg_byte_r+1;
              byte_cnt_r          <= byte_cnt_r+1;

              if byte_cnt_r<unsigned(cmd_nbytes)-1 then 
                if reg_byte_r(1 downto 0)="11" then
                  crx_state_r       <= wrcmd_wbinc_st;

                else 
                  crx_state_r       <= wrcmd_wbmodify_st;

                end if;
              else
                crx_state_r         <= wrcmd_wr_ack_st;

              end if;

            elsif crx_wb_sm.err='1' then
              wb_cyc_r            <= '0';
              wb_wr_r             <= '0';

              if send_wr_ack_g then
                crx_state_r         <= wrcmd_wr_err_st;

              else
                crx_state_r         <= idle_st;
              
              end if;
    
            else
              wb_cyc_r          <= '1';
              wb_wr_r           <= '1';

            end if;



          when wrcmd_wbmodify_st => 
            
            crx_state_r         <= wrcmd_w8wrwb_st;

            regbyte_v           := to_integer(reg_byte_r);

            wb_dat_new_v        := wb_rddat_r;
            wb_dat_new_v(8*(regbyte_v+1)-1 downto 8*regbyte_v)          
                                := cmd_fifo_dat;
            wb_wrdat_r          <= wb_dat_new_v;
            wb_rddat_r          <= wb_dat_new_v;
            cmd_fifo_rd         <= '1';


          when wrcmd_wr_ack_st =>
            rep_fifo_din  <= x"01";
            rep_fifo_wr   <= '1';
            crx_state_r   <= wrcmd_send_st;
          
          when wrcmd_wr_err_st =>
            rep_fifo_din  <= x"02";
            rep_fifo_wr   <= '1';
            crx_state_r   <= wrcmd_send_st;

          when wrcmd_send_st =>
            crx_state_r   <= idle_st;

            rep_send      <= '1';



          when others =>
            null;
        end case;
      end if;
    end if;
  end process fsm;

  emu: entity work.ic_top
  generic map(
    g_FIFO_DEPTH        => 32                         --! Depth of the internal FIFO used to improve the timming performance
  )
  port map(
    -------------------------------------------
    ------- CLock and reset -------------------
    -------------------------------------------
    tx_clk_i            => rx_clk_i,                   --! Tx clock (Tx_frameclk_o from GBT-FPGA IP): must be a multiple of the LHC frequency
    tx_clk_en           => '1',                        --! Tx clock enable signal must be used in case of multi-cycle path(tx_clk_i > LHC frequency). By default: always enabled
  
    rx_clk_i            => rx_clk_i,                   --! Rx clock (Rx_frameclk_o from GBT-FPGA IP): must be a multiple of the LHC frequency
    rx_clk_en           => '1',                        --! Rx clock enable signal must be used in case of multi-cycle path(rx_clk_i > LHC frequency). By default: always enabled
  
    rx_reset_i          => rx_reset,                 --! Reset RX datapath
    tx_reset_i          => rx_reset,                 --! Reset TX datapath
    
    -------------------------------------------
    ------- GBTxEMU transmit ------------------
    ---- Replies to Master commands -----------
    -------------------------------------------
    
    --! I2C address of the GBTxEMU
    tx_GBTx_address_i   => emu_id_g,                

    --! GBTxEMU is a slave. It should only reply with data if the master requests a read
    --! Not sure what should be in the frame
    --! The data from command are as good as any
    tx_register_addr_i  => cmd_byteaddr,                   

    --! GBTxEMU is a slave. It should only reply with data if the master requests a read
    --! Not sure what should be in the frame
    --! The data from command are as good as any
    tx_nb_to_be_read_i  => cmd_nbytes,
    

    --! Request a write operation into the internal FIFO
    tx_wr_i             => rep_fifo_wr,

    --! Data to be written into the internal FIFO
    tx_data_to_gbtx_i   => rep_fifo_din,                

    --! GBTxEMU doesn't respond for write commands
    tx_start_write_i    => rep_send,

    --! Trigger the sending of requested data
    tx_start_read_i     => '0',       

    --! IC core ready for a transaction
    tx_ready_o          => rep_fifo_ready,
    

    -------------------------------------------
    -------- GBTxEMU receive ------------------
    -------- Master commands ------------------
    -------------------------------------------

    --! Command FIFO is empty
    rx_empty_o          => cmd_fifo_empty,

    --! Command to read single word from RX FIFO
    rx_rd_i             => cmd_fifo_rd,                
    --! Command FIFO data output
    --! In the EMU it should only contain data to be written by write command
    rx_data_from_gbtx_o => cmd_fifo_dat,                

    --! Address of the first register to accessed
    rx_mem_ptr_o        => cmd_byteaddr,                 

    --! Number of words/bytes to be read/written
    rx_nb_of_words_o    => cmd_nbytes,                 

    --! 1 means that read command was received
    --! 0 means that write command was received
    rx_rd_cmd_o         => cmd_rd,                      

    --! I2C address of the GBTx device that should execute the command
    rx_gbtx_addr_o      => cmd_gbtid, 
    rx_parity_err_o     => cmd_parity_error,                

    -------------------------------------------
    -------- Serialized data ------------------
    -------------------------------------------

    tx_data_o           => tx_data_o,             --! (TX) Array of bits to be mapped to the TX GBT-Frame (bits 83/84)
    rx_data_i           => rx_data_i  
  );


  cdc_rst_n         <= not(wb_rst_i or rx_reset_i or rst_from_protector);

  bus_cdc_inst: entity work.wb_cdc 
  generic map( 
    width => 32
  )
  port map(
    rst_n_i        => cdc_rst_n,
    -- Part connecting to the master (as a slave), working in the first clock domain
    slave_clk_i    => rx_clk_i,
    slave_i        => crx_wb_ms,
    slave_o        => crx_wb_sm,

    -- Part connecting to slaves (as the master) working in the second clock domain
    master_clk_i   => wb_clk_i,
    master_i       => wb_sm_i,
    master_o       => wb_ms_o
  );

  fault_protector : process (rx_clk_i)
  begin
    if (rising_edge(rx_clk_i)) then
      if rx_reset_i='1' then
        rst_from_protector  <= '1';
        protector_cnt       <= c_protector_max;
      else
      
        if rx_data_uncertain_i='1' then
          rst_from_protector  <= '1';
          protector_cnt       <= c_protector_max;

        else
          if rx_data_i="11" then
            if protector_cnt>0 then
              protector_cnt <= protector_cnt-1;
            else
              rst_from_protector <= '0';
            end if;
          end if;
        end if;

      end if;
    end if;
  end process fault_protector;

end  ic_wb_ctrl_arc;