library ieee;
  use ieee.std_logic_1164.all;
  use ieee.numeric_std.all;               -- for type conversions

library general_cores;
  use general_cores.wishbone_pkg.all;
  
library work;
  use work.dummy_register_wb_pkg.all;

entity tb_ic is
end tb_ic;

architecture tb_ic_arc of tb_ic is
  
  signal  clk_half_period_c : time := 5 ns;

  signal  clk         : std_logic := '0';
  signal  rst         : std_logic := '0';
  signal  rst_n       : std_logic;

  signal  dpb_wrclken     : std_logic;
  signal  dpb_rdclken     : std_logic;
  signal  dpb_gbtid     : std_logic_vector(6 downto 0);
  signal  dpb_regaddr   : std_logic_vector(15 downto 0);    --! Address of the first register to be accessed
  signal  dpb_rdregnr   : std_logic_vector(15 downto 0);    --! Number of words/bytes to be read (only for read transactions)
  signal  dpb_wrdat_en  : std_logic;
  signal  dpb_wrdat     : std_logic_vector(7 downto 0);
  signal  dpb_rddat_en  : std_logic;
  signal  dpb_rddat     : std_logic_vector(7 downto 0);
  signal  dpb_issue_wrcmd  : std_logic;
  signal  dpb_issue_rdcmd  : std_logic;
  signal  dpb_ready_for_write  : std_logic;
  signal  dpb_rdfifo_empty  : std_logic;
        
       
  signal  serial_dpb_to_emu     : std_logic_vector(1 downto 0);
  signal  serial_dpb_to_emu_with_errors     : std_logic_vector(1 downto 0);
  signal  serial_emu_to_dpb     : std_logic_vector(1 downto 0);

  signal wb_ms        : t_wishbone_master_out;
  signal wb_sm        : t_wishbone_slave_out;
  
  signal clk_tx       : std_logic;
  signal clk_rx       : std_logic;
  signal crx_rst      : std_logic;
  signal ctx_rst      : std_logic;

  signal break_transmission      : std_logic;

begin

  DPB: entity work.ic_top
  generic map(
    g_FIFO_DEPTH        => 32                         --! Depth of the internal FIFO used to improve the timming performance
  )
  port map(
    -- Clock and reset
    tx_clk_i            => clk,                        --! Tx clock (Tx_frameclk_o from GBT-FPGA IP): must be a multiple of the LHC frequency
    tx_clk_en           => dpb_wrclken,                        --! Tx clock enable signal must be used in case of multi-cycle path(tx_clk_i > LHC frequency). By default: always enabled
  
    rx_clk_i            => clk,                        --! Rx clock (Rx_frameclk_o from GBT-FPGA IP): must be a multiple of the LHC frequency
    rx_clk_en           => dpb_rdclken,                        --! Rx clock enable signal must be used in case of multi-cycle path(rx_clk_i > LHC frequency). By default: always enabled
  
    rx_reset_i          => rst,                        --! Reset RX datapath
    tx_reset_i          => rst,                        --! Reset TX datapath
    
    -- Configuration
    tx_GBTx_address_i   => dpb_gbtid,                   --! I2C address of the GBTx
    tx_register_addr_i  => dpb_regaddr,                 --! Address of the first register to be accessed
    tx_nb_to_be_read_i  => dpb_rdregnr,                 --! Number of words/bytes to be read (only for read transactions)
    
    -- Internal FIFO
    tx_wr_i             => dpb_wrdat_en,                        --! Request a write operation into the internal FIFO (Data to GBTx)
    tx_data_to_gbtx_i   => dpb_wrdat,                     --! Data to be written into the internal FIFO
    
    rx_rd_i             => dpb_rddat_en,                        --! Request a read operation of the internal FIFO (GBTx reply)
    rx_data_from_gbtx_o => dpb_rddat,             --! Data from the FIFO
    
    -- FSM Control
    tx_start_write_i    => dpb_issue_wrcmd,                        --! Request a write config. to the GBTx
    tx_start_read_i     => dpb_issue_rdcmd,                        --! Request a read config. to the GBTx
  
    -- Status
    tx_ready_o          => dpb_ready_for_write,                        --! IC core ready for a transaction
    rx_empty_o          => dpb_rdfifo_empty,                       --! Rx FIFO is empty (no reply from GBTx)
    
    rx_gbtx_addr_o      => open,                        --! I2C address of the GBTx (read from a reply)
    rx_mem_ptr_o        => open,    --! I2C address of the first register read/written
    rx_nb_of_words_o    => open,    --! Number of words/bytes read/written
    
    -- IC lines
    tx_data_o           => serial_dpb_to_emu,     --! (TX) Array of bits to be mapped to the TX GBT-Frame (bits 83/84)
    rx_data_i           => serial_emu_to_dpb  
  );

  ctx_rst   <= rst;
  crx_rst   <= rst;
  clk_tx    <= clk;
  clk_rx    <= clk;

  emu: entity work.ic_wb_ctrl
    generic map(
        emu_id_g     => std_logic_vector(to_unsigned(16#3a#, 7))
      )
    port map(
      --! Wishbone CLK
      wb_clk_i          => clk,
      --! Reset signal synchronous with Wishbone CLK
      wb_rst_i          => rst,

      --! Wishbone output record
      wb_ms_o           => wb_ms,
      --! Wishbone input record
      wb_sm_i           => wb_sm,               
    
      --! Rx clock (Rx_frameclk_o from GBT-FPGA IP): must be a multiple of the LHC frequency
      rx_clk_i          => clk_rx,              
      --! Rx clock enable signal must be used in case of multi-cycle path(rx_clk_i > LHC frequency). By default: always enabled
      rx_clk_en         => '1',
    
      --! Reset RX datapath
      rx_reset_i          => crx_rst,                     
       
      rx_data_uncertain_i   => break_transmission,

      --! (TX) Array of bits to be mapped to the TX GBT-Frame (bits 83/84)
      tx_data_o           => serial_emu_to_dpb,     
      --! (RX) Array of bits to be mapped to the RX GBT-Frame (bits 83/84)
      rx_data_i           => serial_dpb_to_emu_with_errors
  );

  serial_dpb_to_emu_with_errors <= serial_dpb_to_emu when break_transmission='0' else not serial_dpb_to_emu;

  test_reg: entity work.dummy_register_wb
  port map(
    slave_i               => wb_ms,
    slave_o               => wb_sm,

    rst_n_i               => rst_n,
    clk_sys_i             => clk
  );

  clkgen : process 
  begin
    clk     <= not clk;
    wait for clk_half_period_c;
  end process clkgen;

  control_ap : process
    
    procedure wait_clk (nr : natural) is
    begin
      for i in 0 to nr-1 loop
        wait until clk='0';
        wait until clk='1';
      end loop;
    end wait_clk;

    procedure dpb_write_reg( addr: natural;
                              val: unsigned;
                              bnr: natural )  is
      variable tmp : std_logic_vector( 160-1 downto 0);

    begin
 
        dpb_regaddr       <= std_logic_vector(to_unsigned(addr, 16));
        tmp               := std_logic_vector(resize(val, 160));
        identifier : for i in 0 to bnr-1 loop
          dpb_wrdat         <= tmp(8*(i+1)-1 downto 8*i);
          dpb_wrdat_en      <= '1';
          wait_clk(1);
          dpb_wrdat_en      <= '0';
          
        end loop;

        dpb_issue_wrcmd   <= '1';
        wait_clk(1);
        dpb_issue_wrcmd   <= '0';

    end dpb_write_reg;

    procedure dpb_read_reg( addr: natural; nbytes: natural )  is
    begin
 
        dpb_regaddr       <= std_logic_vector(to_unsigned(addr, 16));
        dpb_rdregnr       <= std_logic_vector(to_unsigned(nbytes, 16));

        dpb_issue_rdcmd   <= '1';
        wait_clk(1);
        dpb_issue_rdcmd   <= '0';

    end dpb_read_reg;

  begin
    dpb_gbtid           <= "0101010";
    dpb_wrclken         <= '1';
    dpb_rdclken         <= '1';

    break_transmission <= '0';

    dpb_issue_wrcmd   <= '0';
    dpb_issue_rdcmd   <= '0';
    
    rst       <= '0';
    wait_clk(4);
    rst       <= '1';
    wait_clk(4);
    rst       <= '0';
    wait_clk(50);

    dpb_read_reg(0, 4);
    wait until dpb_rdfifo_empty='0';
    dpb_rddat_en  <= '1';
    wait_clk(2);

    --register 0 has constant value
    assert dpb_rddat=x"33" report "Incorrect data read!" severity error;
    wait_clk(1);
    assert dpb_rddat=x"bf" report "Incorrect data read!" severity error;
    wait_clk(1);
    assert dpb_rddat=x"c0" report "Incorrect data read!" severity error;
    wait_clk(1);
    assert dpb_rddat=x"53" report "Incorrect data read!" severity error;
    wait_clk(1);
    assert dpb_rdfifo_empty='1' report "FIFO should be empty!" severity error;
    dpb_rddat_en  <= '0';
    
    dpb_read_reg(4, 8);
    wait until dpb_rdfifo_empty='0';
    dpb_rddat_en  <= '1';
    wait_clk(2);
    
    --register 2 has constant value
    assert dpb_rddat=x"31" report "Incorrect data read!" severity error;
    wait_clk(1);
    assert dpb_rddat=x"5d" report "Incorrect data read!" severity error;
    wait_clk(1);
    assert dpb_rddat=x"b2" report "Incorrect data read!" severity error;
    wait_clk(1);
    assert dpb_rddat=x"5c" report "Incorrect data read!" severity error;
    
    --register 3 has default value of 0
    wait_clk(1);
    assert dpb_rddat=x"00" report "Incorrect data read!" severity error;
    wait_clk(1);
    assert dpb_rddat=x"00" report "Incorrect data read!" severity error;
    wait_clk(1);
    assert dpb_rddat=x"00" report "Incorrect data read!" severity error;
    wait_clk(1);
    assert dpb_rddat=x"00" report "Incorrect data read!" severity error;
    
    wait_clk(1);
    assert dpb_rdfifo_empty='1' report "FIFO should be empty!" severity error;
    dpb_rddat_en  <= '0';

    dpb_read_reg(2, 4);
    wait until dpb_rdfifo_empty='0';
    dpb_rddat_en  <= '1';
    
    --second half of reg 0 
    wait_clk(2);
    assert dpb_rddat=x"c0" report "Incorrect data read!" severity error;
    wait_clk(1);
    assert dpb_rddat=x"53" report "Incorrect data read!" severity error;
    
    --first half of reg 1 
    wait_clk(1);
    assert dpb_rddat=x"31" report "Incorrect data read!" severity error;
    wait_clk(1);
    assert dpb_rddat=x"5d" report "Incorrect data read!" severity error;

    wait_clk(1);
    assert dpb_rdfifo_empty='1' report "FIFO should be empty!" severity error;
    dpb_rddat_en  <= '0';
  
    --write 1 byte (0xef) to register 5
    dpb_write_reg(5*4, x"123456789abcdef", 1);
    -- wait for ack
    wait until dpb_rdfifo_empty='0';
    dpb_rddat_en  <= '1';
    wait_clk(2);
    --check if received byte is ack
    assert dpb_rddat=x"01" report "Not an ACK!" severity error;
    wait_clk(1);
    assert dpb_rdfifo_empty='1' report "FIFO should be empty!" severity error;
    dpb_rddat_en  <= '0';

    --write 3 bytes of 0x0 to register 5
    dpb_write_reg(5*4, x"0", 3);
    -- wait for ack
    wait until dpb_rdfifo_empty='0';
    dpb_rddat_en  <= '1';
    wait_clk(2);
    --check if received byte is ack
    assert dpb_rddat=x"01" report "Not an ACK!" severity error;
    wait_clk(1);
    assert dpb_rdfifo_empty='1' report "FIFO should be empty!" severity error;
    dpb_rddat_en  <= '0';

    --write 6 bytes to register 5 (and 6)
    dpb_write_reg(5*4, x"123456789abcdef", 6);
    -- wait for ack
    wait until dpb_rdfifo_empty='0';
    dpb_rddat_en  <= '1';
    wait_clk(2);
    --check if received byte is ack
    assert dpb_rddat=x"01" report "Not an ACK!" severity error;
    wait_clk(1);
    assert dpb_rdfifo_empty='1' report "FIFO should be empty!" severity error;
    dpb_rddat_en  <= '0';

    --read back
    dpb_read_reg(5*4, 8);
    wait until dpb_rdfifo_empty='0';
    dpb_rddat_en  <= '1';
    wait_clk(2);
    assert dpb_rddat=x"ef" report "Invalid readback data!" severity error;
    wait_clk(1);
    assert dpb_rddat=x"cd" report "Invalid readback data!" severity error;
    wait_clk(1);
    assert dpb_rddat=x"ab" report "Invalid readback data!" severity error;
    wait_clk(1);
    assert dpb_rddat=x"89" report "Invalid readback data!" severity error;
    wait_clk(1);
    assert dpb_rddat=x"67" report "Invalid readback data!" severity error;
    wait_clk(1);
    assert dpb_rddat=x"45" report "Invalid readback data!" severity error;
    wait_clk(1);
    assert dpb_rddat=x"00" report "Invalid readback data!" severity error;
    wait_clk(1);
    assert dpb_rddat=x"00" report "Invalid readback data!" severity error;
    wait_clk(1);
    assert dpb_rdfifo_empty='1' report "FIFO should be empty!" severity error;
    dpb_rddat_en  <= '0';

    dpb_write_reg(5*4+1, x"123123123123123", 7);
    wait until dpb_rdfifo_empty='0';
    dpb_rddat_en  <= '1';
    wait until dpb_rdfifo_empty='1';
    dpb_rddat_en  <= '0';


    dpb_read_reg(23, 4);
    wait until dpb_rdfifo_empty='0';
    dpb_rddat_en  <= '1';
    wait until dpb_rdfifo_empty='1';
    dpb_rddat_en  <= '0';

    ---------------------------------------------
    ----Reproducing a bug found in hardware ------
    ---------------------------------------------

    ----- these two commands worked fine -----------
    dpb_write_reg(16#0D#*4, x"000000c7", 4);
    wait until dpb_rdfifo_empty='0';
    dpb_rddat_en  <= '1';
    wait until dpb_rdfifo_empty='1';
    dpb_rddat_en  <= '0';

    --dpb_read_reg(16#48#, 4);
    --wait until dpb_rdfifo_empty='0';
    --dpb_rddat_en  <= '1';
    --wait until dpb_rdfifo_empty='1';
    --dpb_rddat_en  <= '0';

    ---------------------------------------------

    ------------- write worked fine
    ------------- but instead of following read a write with garbage parameters was issued by IC iface

    dpb_write_reg(16#0D#*4, x"000000c8", 4);
    wait until dpb_rdfifo_empty='0';
    dpb_rddat_en  <= '1';
    wait until dpb_rdfifo_empty='1';
    dpb_rddat_en  <= '0';

    dpb_read_reg(16#48#, 4);
    wait until dpb_rdfifo_empty='0';
    dpb_rddat_en  <= '1';
    wait until dpb_rdfifo_empty='1';
    dpb_rddat_en  <= '0';

    -------------------------------------------------
    -- check what happens when transmission is interrupted
    -------------------------------------------------

    --first correct transaction so it was easier to compare
    dpb_read_reg(0, 4);
    wait until dpb_rdfifo_empty='0';
    dpb_rddat_en  <= '1';
    wait until dpb_rdfifo_empty='1';
    dpb_rddat_en  <= '0';

    wait_clk(3);

    -- now the broken one
    dpb_read_reg(0, 4);
    -- let the emu receive some data
    wait_clk(16);
    -- and then make serial lines stuck at "11" for some time
    break_transmission <= '1';
    wait_clk(4);
    -- revert transmission
    break_transmission <= '0';
    
    -- wait for quite long
    wait_clk(400);

    -- make sure that emu didn't respond
    assert dpb_rdfifo_empty='1' report "EMU shouldn't respond to broken transmission!" severity error;

    -- make sure that it will work again
    dpb_read_reg(0, 4);
    wait until dpb_rdfifo_empty='0';
    dpb_rddat_en  <= '1';
    wait_clk(2);

    --register 0 has constant value
    assert dpb_rddat=x"33" report "Incorrect data read!" severity error;
    wait_clk(1);
    assert dpb_rddat=x"bf" report "Incorrect data read!" severity error;
    wait_clk(1);
    assert dpb_rddat=x"c0" report "Incorrect data read!" severity error;
    wait_clk(1);
    assert dpb_rddat=x"53" report "Incorrect data read!" severity error;
    wait_clk(1);
    assert dpb_rdfifo_empty='1' report "FIFO should be empty!" severity error;
    dpb_rddat_en  <= '0';
   
    wait for 100000 ms;

  end process control_ap;

  rst_n   <= not rst;


end  tb_ic_arc;