library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;               -- for type conversions

library agwb;

use agwb.ec_ctrl_pkg.all;
use agwb.ic_ctrl_pkg.all;

library general_cores;
use general_cores.wishbone_pkg.all;

entity xwb_gbtsc_top is
  port (
    -- EC line
    ec_data_o           : out std_logic_vector(1 downto 0);           --! (TX) Array of bits to be mapped to the TX GBT-Frame
    ec_data_i           : in  std_logic_vector(1 downto 0);           --! (RX) Array of bits to be mapped to the RX GBT-Frame

    -- IC lines
    ic_data_o           : out std_logic_vector(1 downto 0);                 --! (TX) Array of bits to be mapped to the TX GBT-Frame (bits 83/84)
    ic_data_i           : in  std_logic_vector(1 downto 0);                  --! (RX) Array of bits to be mapped to the RX GBT-Frame (bits 83/84);

    slave_i : in t_wishbone_slave_in;
    slave_o : out t_wishbone_slave_out;
    
    rst_n_i : in std_logic;
    clk_sys_i : in std_logic

  ) ;
end xwb_gbtsc_top;

architecture xwb_gbtsc_top_arc of xwb_gbtsc_top is


    signal wb_ic_ms : t_wishbone_master_out;
    signal wb_ic_sm : t_wishbone_master_in;

    signal wb_ec_ms  : t_wishbone_master_out;
    signal wb_ec_sm  : t_wishbone_master_in;
    
    -- -- Signals for EC
    signal ec_ctrl_o         : t_ctrl_ec;
    signal ec_ctrl_o_stb     : std_logic;
    signal ec_tx_cmd_o       : t_tx_cmd;
    signal ec_tx_data_o      : t_ec_tx_data;
    signal ec_status_i       : t_status_ec;
    signal ec_status2_i      : t_status2_ec;
    signal ec_rx_data_i      : t_ec_rx_data;
    signal ec_rx_data_i_ack  : std_logic;
    -- Signals for IC
    signal ic_ctrl_o         : t_ctrl_ic;
    signal ic_ctrl_o_stb     : std_logic;
    signal ic_tx_rega_nbtr_o : t_tx_rega_nbtr;
    signal ic_tx_data_o      : t_ic_tx_data;
    signal ic_tx_data_o_stb  : std_logic;
    signal ic_status_i       : t_status_ic;
    signal ic_rx_mptr_nbw_i  : t_rx_mptr_nbw;
    signal ic_rx_data_i      : t_ic_rx_data;
    signal ic_rx_data_i_ack  : std_logic;

    signal rx_received       : std_logic;
    signal rx_received_latch : std_logic;

  begin  -- block gbtsc_wb1
    
    gbtsc_top_inst : entity agwb.gbtsc_top
    port map (
      slave_i   => slave_i,
      slave_o   => slave_o,
      ic_wb_m_o => wb_ic_ms,
      ic_wb_m_i => wb_ic_sm,
      ec_wb_m_o => wb_ec_ms,
      ec_wb_m_i => wb_ec_sm,
      rst_n_i   => rst_n_i,
      clk_sys_i => clk_sys_i
    );

    agwb_ec_ctrl_wb_1 : entity agwb.ec_ctrl
    port map (
      slave_i       => wb_ec_ms,
      slave_o       => wb_ec_sm,
      ctrl_o        => ec_ctrl_o,
      ctrl_o_stb    => ec_ctrl_o_stb,
      tx_cmd_o      => ec_tx_cmd_o,
      tx_data_o     => ec_tx_data_o,
      status_i      => ec_status_i,
      status2_i     => ec_status2_i,
      rx_data_i     => ec_rx_data_i,
      rx_data_i_ack => ec_rx_data_i_ack,
      rst_n_i       => rst_n_i,
      clk_sys_i     => clk_sys_i);

    agwb_ic_ctrl_wb_1 : entity agwb.ic_ctrl
    port map (
      slave_i        => wb_ic_ms,
      slave_o        => wb_ic_sm,
      ctrl_o         => ic_ctrl_o,
      ctrl_o_stb     => ic_ctrl_o_stb,
      tx_rega_nbtr_o => ic_tx_rega_nbtr_o,
      tx_data_o      => ic_tx_data_o,
      tx_data_o_stb  => ic_tx_data_o_stb,
      status_i       => ic_status_i,
      rx_mptr_nbw_i  => ic_rx_mptr_nbw_i,
      rx_data_i      => ic_rx_data_i,
      rx_data_i_ack  => ic_rx_data_i_ack,
      rst_n_i        => rst_n_i,
      clk_sys_i      => clk_sys_i);

    gbtsc : entity work.gbtsc_top
    generic map (
      -- IC configuration
      g_IC_FIFO_DEPTH => 20,
      -- EC configuration
      g_SCA_COUNT     => 1
      )
    port map (
      -- Clock & reset
      tx_clk_i            => clk_sys_i,
      tx_clk_en           => '1',
      rx_clk_i            => clk_sys_i,
      rx_clk_en           => '1',
      rx_reset_i          => ic_ctrl_o.reset(0),
      tx_reset_i          => ic_ctrl_o.reset(0),
      -- IC control
      tx_start_write_i    => ic_ctrl_o.start_write(0) and ic_ctrl_o_stb,
      tx_start_read_i     => ic_ctrl_o.start_read(0) and ic_ctrl_o_stb,
      -- IC configuration
      tx_GBTx_address_i   => ic_ctrl_o.gbtx_addr,
      tx_register_addr_i  => ic_tx_rega_nbtr_o.reg_addr,
      tx_nb_to_be_read_i  => ic_tx_rega_nbtr_o.bytes_to_read,
      -- IC FIFO control
      tx_wr_i             => ic_tx_data_o_stb,
      tx_data_to_gbtx_i   => ic_tx_data_o,
      rx_rd_i             => ic_rx_data_i_ack,
      rx_data_from_gbtx_o => ic_rx_data_i,
      -- IC Status
      tx_ready_o          => ic_status_i.tx_ready(0),
      rx_empty_o          => ic_status_i.rx_empty(0),
      rx_gbtx_addr_o      => ic_status_i.gbtx_addr,
      rx_mem_ptr_o        => ic_rx_mptr_nbw_i.mem_ptr,
      rx_nb_of_words_o    => ic_rx_mptr_nbw_i.words_read,
      -- SCA control
      sca_enable_i        => "1",
      start_reset_cmd_i   => ec_ctrl_o.start_reset(0) and ec_ctrl_o_stb,
      start_connect_cmd_i => ec_ctrl_o.start_connect(0) and ec_ctrl_o_stb,
      start_command_i     => ec_ctrl_o.start_cmd(0) and ec_ctrl_o_stb,
      inject_crc_error    => ec_ctrl_o.inject_crc_error(0) and ec_ctrl_o_stb,
      -- SCA command
      tx_address_i        => ec_tx_cmd_o.addr,
      tx_transID_i        => ec_tx_cmd_o.trans_id,
      tx_channel_i        => ec_tx_cmd_o.chan,
      tx_command_i        => ec_tx_cmd_o.command,
      tx_data_i           => ec_tx_data_o,
      rx_received_o(0)    => rx_received,
      rx_address_o(0)     => ec_status2_i.addr,
      rx_control_o(0)     => ec_status2_i.control,
      rx_transID_o(0)     => ec_status2_i.trans_id,
      rx_channel_o(0)     => ec_status2_i.chan,
      rx_len_o(0)         => ec_status_i.len,
      rx_error_o(0)       => ec_status_i.error,
      rx_data_o(0)        => ec_rx_data_i,
      -- EC line
      ec_data_o(0)        => ec_data_o,
      ec_data_i(0)        => ec_data_i,
      -- IC lines
      ic_data_o           => ic_data_o,
      ic_data_i           => ic_data_i
      );

    ec_reply_latch : process(clk_sys_i) is
    begin
      if rising_edge(clk_sys_i) then
        -- clear the flag either on a) start of new command b) readout of command response
        if ic_ctrl_o.reset(0) = '1' or ec_ctrl_o.start_cmd(0) = '1' or ec_rx_data_i_ack = '1' then
          rx_received_latch <= '0';
        elsif rx_received = '1' then
          rx_received_latch <= '1';
        end if;
      end if;
    end process;

    ec_status_i.received(0) <= rx_received_latch;


end  xwb_gbtsc_top_arc;