library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;               -- for type conversions

library general_cores;
use general_cores.wishbone_pkg.all;

use work.gbt_fpga_common_pkg.all;
use work.gbt_bank_package.all;

library agwb;

entity crob_interface is
  generic(
    links_g : natural:= 3
  );
  port (
    -- experiment reference clock
    -- must be 40 MHz
    -- also used for wishbone
    clk_frame_i          : in  std_logic;

    -- reference clock for the MGT
    -- must be 120 MHz
    -- must come from ibuf (not bufg)
    -- MUST BE SYNCHRONOUS WITH clk_frame_i!!!
    clk_mgtref_i         : in  std_logic;

    --==============--
    -- Reset        --
    --==============--
    -- reset_async_i is synchronized to clk_frame_i inside the core
    -- asserting reset_async_i will set a reset signal accessible via wishbone
    -- after toggling reset_async_i user must make sure that clk_frame_i and clk_mgtref_i are stable
    -- and remove reset via gbt_fpga_common ctrl.rst register
    reset_async_i       : in std_logic;

    --==============--
    -- Serial lanes --
    --==============--

    -- serial link connected to master GBTx
    serial_master_downlink_o  : out t_diff_pair;
    serial_master_uplink_i    : in  t_diff_pair;
    
    -- serial links connected to slave GBTx's (if there are any)
    serial_slave0_downlink_o  : out t_diff_pair;
    serial_slave0_uplink_i    : in  t_diff_pair := ('0', '0');
    
    serial_slave1_downlink_o  : out t_diff_pair;
    serial_slave1_uplink_i    : in  t_diff_pair := ('0', '0');

    --==============--
    -- Data          --
    --==============--  
    
    -- data to be sent via master downlink (no EC/IC)
    downlink_i                : in  std_logic_vector(79 downto 0);
    -- data received via all (master and slave) uplinks (no EC/IC)
    uplink_master_o          : out std_logic_vector(112-1 downto 0);
    uplink_slave0_o          : out std_logic_vector(112-1 downto 0);
    uplink_slave1_o          : out std_logic_vector(112-1 downto 0);

    --==============--
    -- Wishbone slave interface      --
    --==============--      
    wb_rst_i            : in std_logic;
    wbs_i               : in t_wishbone_slave_in;
    wbs_o               : out t_wishbone_slave_out

  ) ;
end crob_interface;

architecture crob_interface_arc of crob_interface is

  signal GBTBANK_MGT_RX_P : std_logic_vector(links_g-1 downto 0);
  signal GBTBANK_MGT_RX_N : std_logic_vector(links_g-1 downto 0);
  signal GBTBANK_MGT_TX_P : std_logic_vector(links_g-1 downto 0);
  signal GBTBANK_MGT_TX_N : std_logic_vector(links_g-1 downto 0);

  --signal uplinks         : t_slv116_arr(links_g-1 downto 0);
  --signal downlinks       : t_slv84_arr(links_g-1 downto 0);
  signal uplinks         : gbt_reg116_A(links_g-1 downto 0);
  signal downlinks       : gbt_reg84_A(links_g-1 downto 0);

  signal downlink_ic     : std_logic_vector(1 downto 0);
  signal downlink_ec     : std_logic_vector(1 downto 0);

  signal uplink_ic       : std_logic_vector(1 downto 0);
  signal uplink_ec       : std_logic_vector(1 downto 0);

  signal wb_gbtsc_ms     : t_wishbone_slave_in;
  signal wb_gbtsc_sm     : t_wishbone_slave_out;

  signal wb_gbtcommon_ms : t_wishbone_slave_in;
  signal wb_gbtcommon_sm : t_wishbone_slave_out;

  signal wb_gbtlink_ms   : t_wishbone_master_out_array(links_g-1 downto 0);
  signal wb_gbtlink_sm   : t_wishbone_master_in_array(links_g-1 downto 0);

begin
  
  assert (links_g=1) or (links_g=3) report "crob_interface: generic links_g must be 1 or 3!" severity FAILURE;
  
  ------------------------------------------------------------------------------
  -- serial lines mapping
  ------------------------------------------------------------------------------

  --connect 0'th RX/TX pair to master link
  serial_master_downlink_o.p   <= GBTBANK_MGT_TX_P(0);
  serial_master_downlink_o.n   <= GBTBANK_MGT_TX_N(0);
  GBTBANK_MGT_RX_P(0)          <= serial_master_uplink_i.p; 
  GBTBANK_MGT_RX_N(0)          <= serial_master_uplink_i.n;

  -- others are slaves
  serial_slave_mapping_gen : if links_g=3 generate
    serial_slave0_downlink_o.p  <= GBTBANK_MGT_TX_P(1);
    serial_slave0_downlink_o.n  <= GBTBANK_MGT_TX_N(1);
    GBTBANK_MGT_RX_P(1)         <= serial_slave0_uplink_i.p; 
    GBTBANK_MGT_RX_N(1)         <= serial_slave0_uplink_i.n;

    serial_slave1_downlink_o.p  <= GBTBANK_MGT_TX_P(2);
    serial_slave1_downlink_o.n  <= GBTBANK_MGT_TX_N(2);
    GBTBANK_MGT_RX_P(2)         <= serial_slave1_uplink_i.p; 
    GBTBANK_MGT_RX_N(2)         <= serial_slave1_uplink_i.n;
  end generate serial_slave_mapping_gen;

  -----------------------------------------------------------------------------
  -- paralleled data mapping
  ------------------------------------------------------------------------------
  -- downlinks signal is concatenated form data (from input port) and ic/ec data (from SCA core)
  
  -- master link
  downlinks(0)(80-1 downto 0)    <= downlink_i;
  uplink_master_o                <= uplinks(0)(112-1 downto 0);

  -- slave links
  paralleled_data_mapping_slave_gen : if links_g=3 generate
    downlinks(1)(80-1 downto 0)      <= (others => '0');
    downlinks(2)(80-1 downto 0)      <= (others => '0');
    
    uplink_slave0_o   <= uplinks(1)(112-1 downto 0);
    uplink_slave1_o   <= uplinks(2)(112-1 downto 0);
    
  end generate paralleled_data_mapping_slave_gen;

  ------------------------------------------------------------------------------
  -- IC/EC mapping
  ------------------------------------------------------------------------------
  -- downlinks signal is concatenated form data (from input port) and ic/ec data (from sca core)

  -- uplinks work in WBus (wide bus) mode
  -- got to add 32 to the bit index in the downlink frame (normally 80-84)
  uplink_ic <= uplinks(0)(115 downto 114);
  uplink_ec <= uplinks(0)(113 downto 112);

  downlinks(0)(83 downto 82) <= downlink_ic;
  downlinks(0)(81 downto 80) <= downlink_ec;
  --downlinks(1)(83 downto 82) <= downlink_ic;
  --downlinks(1)(81 downto 80) <= downlink_ec;
  --downlinks(2)(83 downto 82) <= downlink_ic;
  --downlinks(2)(81 downto 80) <= downlink_ec;

  icec_mapping_slave_gen : if links_g=3 generate
    downlinks(1)(83 downto 82) <= (others => '0');
    downlinks(1)(81 downto 80) <= (others => '0');
    downlinks(2)(83 downto 82) <= (others => '0');
    downlinks(2)(81 downto 80) <= (others => '0');
  end generate icec_mapping_slave_gen;

  ------------------------------------------------------------------------------
  -- GBT-FPGA bank wrapper
  ------------------------------------------------------------------------------
  gbt_fpga_common_inst : entity work.gbt_fpga_common
    generic map (
      NUM_LINKS         => links_g
    )
    port map (
      frame_clk_i        => clk_frame_i,
      mgt_refclk_i       => clk_mgtref_i,
      reset_async_i      => reset_async_i,

      GBTBANK_MGT_RX_P   => GBTBANK_MGT_RX_P,
      GBTBANK_MGT_RX_N   => GBTBANK_MGT_RX_N,
      GBTBANK_MGT_TX_P   => GBTBANK_MGT_TX_P,
      GBTBANK_MGT_TX_N   => GBTBANK_MGT_TX_N,
      GBTBANK_GBT_DATA_I => downlinks,
      GBTBANK_WB_DATA_O  => uplinks,
      wb_gbtcommon_i     => wb_gbtcommon_ms,
      wb_gbtcommon_o     => wb_gbtcommon_sm,
      wb_gbtlink_i       => wb_gbtlink_ms,
      wb_gbtlink_o       => wb_gbtlink_sm
    );  
  
  -- ------------------------------------------------------------------------------
  -- GBT Slow Control wrapper
  ------------------------------------------------------------------------------
  xwb_gbtsc_inst : entity work.xwb_gbtsc_top
    port map (
      ec_data_o => downlink_ec,
      ec_data_i => uplink_ec,
      ic_data_o => downlink_ic,
      ic_data_i => uplink_ic,
      slave_i   => wb_gbtsc_ms,
      slave_o   => wb_gbtsc_sm,
      rst_n_i   => not wb_rst_i,
      clk_sys_i => clk_frame_i
    );  

    ----------------------------------------------------------------------------
    -- register definitions (for CROB1 and CROB3)
    ----------------------------------------------------------------------------
    crob3_gen : if links_g=3 generate
      regs_inst : entity agwb.crob3_gbt_iface
        port map (
          slave_i                => wbs_i,
          slave_o                => wbs_o,
          gbtsc_top_wb_m_o       => wb_gbtsc_ms,
          gbtsc_top_wb_m_i       => wb_gbtsc_sm,
          link_wb_m_o            => wb_gbtlink_ms,
          link_wb_m_i            => wb_gbtlink_sm,
          gbt_fpga_common_wb_m_o => wb_gbtcommon_ms,
          gbt_fpga_common_wb_m_i => wb_gbtcommon_sm,
          rst_n_i                => not wb_rst_i,
          clk_sys_i              => clk_frame_i
        );        
    end generate crob3_gen;

    crob1_gen : if links_g=1 generate
      regs_inst : entity agwb.crob1_gbt_iface
        port map (
          slave_i                => wbs_i,
          slave_o                => wbs_o,
          gbtsc_top_wb_m_o       => wb_gbtsc_ms,
          gbtsc_top_wb_m_i       => wb_gbtsc_sm,
          link_wb_m_o            => wb_gbtlink_ms,
          link_wb_m_i            => wb_gbtlink_sm,
          gbt_fpga_common_wb_m_o => wb_gbtcommon_ms,
          gbt_fpga_common_wb_m_i => wb_gbtcommon_sm,
          rst_n_i                => not wb_rst_i,
          clk_sys_i              => clk_frame_i
        );        
    end generate crob1_gen;


end  crob_interface_arc;