-- IEEE VHDL standard library:
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

-- Xilinx devices library:
library unisim;
use unisim.vcomponents.all;

-- Custom libraries and packages:
use work.gbt_bank_package.all;
use work.vendor_specific_gbt_bank_package.all;
use work.gbt_exampledesign_package.all;

library general_cores;
use general_cores.wishbone_pkg.all;


library vhdl_simple;
library agwb;
use agwb.gbt_fpga_link_pkg.all;
use agwb.gbt_fpga_link_pkg;
use agwb.gbt_fpga_common_pkg.all;

--=================================================================================================--
--#######################################   Entity   ##############################################--
--=================================================================================================--


entity gbt_fpga_common is
  generic (
    NUM_LINKS       : integer              := 1;
    g_debug         : string            := "False"
    
  );
  port (

    --==============--
    -- Clocks       --
    --==============--

    -- experiment reference clock
    -- must be 40 MHz
    frame_clk_i          : in  std_logic;

    -- reference clock for the MGT
    -- must be 120 MHz
    -- MUST BE SYNCHRONOUS WITH frame_clk_i!!!
    mgt_refclk_i         : in  std_logic;


    --==============--
    -- Reset        --
    --==============--
    -- reset_async_i is synchronized to frame_clk_i inside the core
    -- asserting reset_async_i will set a reset signal accessible via wishbone
    -- after toggling reset_async_i user must make sure that frame_clk_i and mgt_refclk_i are stable
    -- and remove reset via gbt_fpga_common ctrl.rst register
    reset_async_i       : in std_logic;


    --==============--
    -- Serial lanes --
    --==============--
    GBTBANK_MGT_RX_P : in  std_logic_vector(NUM_LINKS-1 downto 0);
    GBTBANK_MGT_RX_N : in  std_logic_vector(NUM_LINKS-1 downto 0);
    GBTBANK_MGT_TX_P : out std_logic_vector(NUM_LINKS-1 downto 0);
    GBTBANK_MGT_TX_N : out std_logic_vector(NUM_LINKS-1 downto 0);

    --==============--
    -- Data          --
    --==============--      
    GBTBANK_GBT_DATA_I : in  gbt_reg84_A(NUM_LINKS-1 downto 0) := (others => (others => '0'));
    GBTBANK_WB_DATA_I  : in  gbt_reg116_A(NUM_LINKS-1 downto 0) := (others => (others => '0'));
    GBTBANK_GBT_DATA_O : out gbt_reg84_A(NUM_LINKS-1 downto 0);
    GBTBANK_WB_DATA_O  : out gbt_reg116_A(NUM_LINKS-1 downto 0);

    --==============--
    -- Wishbone slave interface      --
    --==============--      
    wb_gbtcommon_i : in t_wishbone_slave_in;
    wb_gbtcommon_o : out t_wishbone_slave_out;
    wb_gbtlink_i   : in t_wishbone_master_out_array(NUM_LINKS-1 downto 0);
    wb_gbtlink_o   : out t_wishbone_master_in_array(NUM_LINKS-1 downto 0)

  );
end gbt_fpga_common;

--=================================================================================================--
--####################################   Architecture   ###########################################-- 
--=================================================================================================--

architecture structural of gbt_fpga_common is

  --================================ Configuration ================================--   

  constant TX_OPTIMIZATION : integer range 0 to 1 := STANDARD;
  constant RX_OPTIMIZATION : integer range 0 to 1 := STANDARD;
  constant TX_ENCODING     : integer range 0 to 2 := GBT_FRAME;
  constant RX_ENCODING     : integer range 0 to 2 := GBT_FRAME;

  -- Extended configuration --
  constant DATA_GENERATOR_ENABLE : integer range 0 to 1 := ENABLED;
  constant CLOCKING_SCHEME       : integer range 0 to 1 := BC_CLOCK;

  --================================ Signal Declarations ================================--   

  --==========--
  -- resets   --
  --==========--
  signal reset_sync              : std_logic;
  signal reset_wb                : std_logic;
  signal gbtbank_general_reset   : std_logic;
  signal gbtbank_manual_reset_rx : std_logic;
  signal gbtbank_manual_reset_tx : std_logic;

  --==========--
  -- GBT Tx   --
  --==========--
  signal gbt_txframeclk_s : std_logic_vector(1 to NUM_LINKS);
  signal gbt_txreset_s    : std_logic_vector(1 to NUM_LINKS);
  signal gbt_txdata_s     : gbt_reg84_A(1 to NUM_LINKS);
  signal wb_txdata_s      : gbt_reg32_A(1 to NUM_LINKS);
  signal gbt_txclken_s    : std_logic_vector(1 to NUM_LINKS);
  signal TX_ENCODING_SEL_i          : std_logic_vector(1 to NUM_LINKS); --! Select the Tx encoding in dynamic mode ('1': GBT / '0': WideBus)


  --==========--
  -- MGT      --
  --==========--
  signal drp_clk        : std_logic;

  signal mgt_txwordclk_s : std_logic_vector(1 to NUM_LINKS);
  signal mgt_rxwordclk_s : std_logic_vector(1 to NUM_LINKS);
  signal mgt_txreset_s   : std_logic_vector(1 to NUM_LINKS);
  signal mgt_rxreset_s   : std_logic_vector(1 to NUM_LINKS);
  signal mgt_txready_s   : std_logic_vector(1 to NUM_LINKS);
  signal mgt_rxready_s   : std_logic_vector(1 to NUM_LINKS);

  signal mgt_headerflag_s       : std_logic_vector(1 to NUM_LINKS);
  signal mgt_devspecific_to_s   : mgtDeviceSpecific_i_R;
  signal mgt_devspecific_from_s : mgtDeviceSpecific_o_R;
  signal resetOnBitslip_s       : std_logic_vector(1 to NUM_LINKS);
  signal mgt_resetoneven        : std_logic_vector(1 to NUM_LINKS);

  --==========--
  -- GBT Rx   --
  --==========--
  signal gbt_rxframeclk_s   : std_logic_vector(1 to NUM_LINKS);
  signal gbt_rxreset_s      : std_logic_vector(1 to NUM_LINKS);
  signal gbt_rxready_s      : std_logic_vector(1 to NUM_LINKS);
  signal gbt_rxdata_s       : gbt_reg84_A(1 to NUM_LINKS);
  signal wb_rxdata_s        : gbt_reg32_A(1 to NUM_LINKS);
  signal gbt_rxclken_s      : std_logic_vector(1 to NUM_LiNKS);
  signal gbt_rxclkenLogic_s : std_logic_vector(1 to NUM_LiNKS);


  --================================--
  -- Data pattern generator/checker --
  --================================--
  signal gbtBank_txEncodingSel         : std_logic_vector(1 downto 0);
  signal gbtBank_rxEncodingSel         : std_logic_vector(1 downto 0);
  signal txData_from_gbtBank_pattGen   : gbt_reg84_A(1 to NUM_LINKS);
  signal txwBData_from_gbtBank_pattGen : gbt_reg32_A(1 to NUM_LINKS);

  signal gbtbank_test_pattern_sel_s          : std_logic_vector(1 downto 0);
  signal gbtbank_test_pattern_databyte_s     : std_logic_vector(7 downto 0);
  signal gbtbank_test_pattern_datawhole_s    : std_logic_vector(79 downto 0);
  signal gbtbank_test_pattern_wbusbyte_s     : std_logic_vector(7 downto 0);
  signal gbtbank_test_pattern_wbuswhole_s    : std_logic_vector(31 downto 0);


  signal GBTBANK_TX_ISDATA_SEL_I             : std_logic_vector(1 to NUM_LINKS);


  signal s_link_ready       : std_logic_vector(1 to NUM_LINKS);
  --==============--
  -- TX Status    --
  --==============--
  signal s_tx_ready      : std_logic_vector(1 to NUM_LINKS);
  signal s_tx_aligned       : std_logic_vector(1 to NUM_LINKS);
  signal s_tx_aligned_val : std_logic_vector(1 to NUM_LINKS);

  --==============--
  -- RX Status    --
  --==============--
  signal gbtbank_gbtrx_ready_s        : std_logic_vector(1 to NUM_LINKS);
  signal gbtbank_gbtrx_ready_r        : std_logic_vector(1 to NUM_LINKS);
  signal gbtbank_gbtrx_ready_loss_s   : std_logic_vector(1 to NUM_LINKS);

  signal s_gbtbank_rx_errordetected               : std_logic_vector(1 to NUM_LINKS);
  signal mgt_rx_header_locked_s                   : std_logic_vector(1 to NUM_LINKS);

  type reg8_a is array (natural range <>) of unsigned(7 downto 0);
  signal s_rx_err_cnt               : reg8_a(1 to NUM_LINKS);

  --==============--
  -- XCVR ctrl    --
  --==============--
  signal GBTBANK_LOOPBACK         : std_logic_vector(2 downto 0);
  signal GBTBANK_TX_POL             : std_logic;
  signal GBTBANK_RX_POL             : std_logic;

  signal pattgen_config         : t_patgen;

  signal wb_links_ms    : t_wishbone_master_out_array(0 to NUM_LINKS-1);
  signal wb_links_sm    : t_wishbone_master_in_array(0 to NUM_LINKS-1);
  
  type t_link_status_arr is array (natural range <>) of gbt_fpga_link_pkg.t_status;
  signal link_statuses  : t_link_status_arr(1 to NUM_LINKS);
  
  type t_link_status_with_clear_arr is array (natural range <>) of gbt_fpga_link_pkg.t_status_with_clear;
  signal link_statuses_cleared      : t_link_status_with_clear_arr(1 to NUM_LINKS);
  signal link_statuses_cleared_stb  : std_logic_vector(1 to NUM_LINKS);

  constant allzeros  : std_logic_vector(1 to NUM_LINKS) := (others => '0');

  attribute MARK_DEBUG : string;
  attribute MARK_DEBUG of gbt_txdata_s : signal is g_debug;
  attribute MARK_DEBUG of gbt_rxdata_s : signal is g_debug;
  attribute MARK_DEBUG of wb_txdata_s : signal is g_debug;
  attribute MARK_DEBUG of wb_rxdata_s : signal is g_debug;

  attribute MARK_DEBUG of gbtbank_general_reset : signal is g_debug;
  attribute MARK_DEBUG of gbtbank_manual_reset_tx : signal is g_debug;
  attribute MARK_DEBUG of gbtbank_manual_reset_rx : signal is g_debug;
  attribute MARK_DEBUG of mgt_txreset_s : signal is g_debug;
  attribute MARK_DEBUG of mgt_rxreset_s : signal is g_debug;
  attribute MARK_DEBUG of gbt_txreset_s : signal is g_debug;
  attribute MARK_DEBUG of gbt_rxreset_s : signal is g_debug;
  attribute MARK_DEBUG of mgt_txready_s : signal is g_debug;
  attribute MARK_DEBUG of mgt_rxready_s : signal is g_debug;
  attribute MARK_DEBUG of reset_sync : signal is g_debug;
  attribute MARK_DEBUG of reset_wb : signal is g_debug;
  
  --=====================================================================================--      

--=================================================================================================--
begin --========####   Architecture Body   ####========-- 
      --=================================================================================================--

  assert (RX_ENCODING/=GBT_DYNAMIC) report "This design was modified in such a way that it doesn't support dynamic encoding!" severity FAILURE;
  assert (TX_ENCODING/=GBT_DYNAMIC) report "This design was modified in such a way that it doesn't support dynamic encoding!" severity FAILURE;

  -- according to gbtxManual
  -- (https://espace.cern.ch/GBT-Project/GBTX/Manuals/gbtxManual.pdf @ 2.2.1) 
  -- the value of the Header field is not used by the GBTx
  -- and we may use it in any way
  -- in the DPB it was always zero
  GBTBANK_TX_ISDATA_SEL_I <= (others => '0');

  mgt_resetoneven <= (others => '0');



  -- in theory it is called a free running clock deeper in the design
  -- frequency must be 40 MHz (set in MGT ipcore)
  -- in my opinion (MG 21.11.2020) it is important that this clock is available during reset of MGTs
  -- it may as well be connected to frame clock
  drp_clk   <= frame_clk_i;
  --==================================== User Logic =====================================--
  

  cpu_reset_sync_i : entity vhdl_simple.reset_synchronizer
    generic map (
      G_INPUT_POSITIVE  => TRUE,
      G_OUTPUT_POSITIVE => TRUE
      )
    port map (
      clk_i  => frame_clk_i,
      rst_i  => reset_async_i,
      rst_o  => reset_sync
    );

  gbtbank_general_reset <= reset_sync or reset_wb;

  --============--
  -- Clocks     --
  --============--              
  gbtBank_Clk_gen : for i in 1 to NUM_LINKS generate

    -- MG (24.08.2020):

    -- this is exactly what gbt_rx_frameclk_phalgnr does in original gbt_fpga_common
    -- when RX is configured to standard mode
    -- it's way more readable this way

    -- use single clock for all uplinks (data from SMX)
    -- we end up with less clock domains and simpler design
    gbt_rxframeclk_s(i) <= frame_clk_i;

    -- in BC mode clken is always 1
    gbt_rxclkenLogic_s(i) <= '1';

    -- in BC mode clken is always 1
    gbt_rxclken_s(i) <= '1';

    -- we don't care about word clocks (so we're not outputting them)
    -- all tx frame clocks must be identical in order to synchronize SMX's (so we're not outputting them)
    -- all rx frame clocks are connected to frame_clk_i in order to simplify design (so we're not outputting them)


    -- this was done in gbt_pattern_generator
    -- moved here in order to improve code readability
    gbt_txframeclk_s(i) <= frame_clk_i;
    gbt_txclken_s(i)    <= '1';
  end generate;

  --============--
  -- Resets     --
  --============--
  gbtBank_rst_gen : for i in 1 to NUM_LINKS generate

    gbtBank_gbtBankRst : entity work.gbt_bank_reset
      generic map (
        INITIAL_DELAY => 1 * 40e6 --          * 1s
      )
      port map (
        GBT_CLK_I     => frame_clk_i,
        TX_FRAMECLK_I => gbt_txframeclk_s(i),
        TX_CLKEN_I    => gbt_txclken_s(i),
        RX_FRAMECLK_I => gbt_rxframeclk_s(i),
        RX_CLKEN_I    => gbt_rxclkenLogic_s(i),
        -- MG (24.08.2020):
        -- the name of the port suggests that we should connect ?mgt_refclk_i?
        -- but signals coming out of this domain seem to be connected to other clock running in drp_clk domain
        MGTCLK_I      => drp_clk,

        --===============--  
        -- Resets scheme --  
        --===============--  
        GENERAL_RESET_I => gbtbank_general_reset,
        TX_RESET_I      => gbtbank_manual_reset_tx,
        RX_RESET_I      => gbtbank_manual_reset_rx,

        MGT_TX_RESET_O => mgt_txreset_s(i),
        MGT_RX_RESET_O => mgt_rxreset_s(i),
        GBT_TX_RESET_O => gbt_txreset_s(i),
        GBT_RX_RESET_O => gbt_rxreset_s(i),

        MGT_TX_RSTDONE_I => mgt_txready_s(i),
        MGT_RX_RSTDONE_I => mgt_rxready_s(i)
      );

    gbtbank_gbtrx_ready_s(i)  <= mgt_rxready_s(i) and gbt_rxready_s(i);

    s_link_ready(i)   <= mgt_txready_s(i) and mgt_rxready_s(i);

    s_tx_ready(i)  <= not(gbt_txreset_s(i));

  end generate;

  --========================--
  -- Data pattern generator --
  --========================--
  dataGenEn_gen : if DATA_GENERATOR_ENABLE = ENABLED generate

    gbtBank_txEncodingSel <= 
      "01" when TX_ENCODING = WIDE_BUS else
      "00" when TX_ENCODING = GBT_FRAME;
 
     gbtBank_rxEncodingSel <= 
      "01" when RX_ENCODING = WIDE_BUS else
      "00" when RX_ENCODING = GBT_FRAME;


    dataGenEn_output_gen : for i in 1 to NUM_LINKS generate
      gbtBank2_pattGen : entity work.gbt_pattern_generator
        generic map(
          CLOCKING_SCHEME => CLOCKING_SCHEME
        )
        port map (
          GENERAL_RST_I => gbtbank_general_reset or gbtbank_manual_reset_tx,
          reset_i       => gbt_txreset_s(i),
          TX_FRAMECLK_I => frame_clk_i,
          TX_WORDCLK_I  => mgt_txwordclk_s(i),

          -- MG (24.08.2020):
          -- in BC mode it was connected to TX_FRAMECLK_I
          -- it's difficult to understand the clocking if assignments are scattered
          -- so it was moved to the "clocking" part of this source file
          TX_FRAMECLK_O => open,
          TX_CLKEN_o    => open,

          -----------------------------------------------     
          TX_ENCODING_SEL_I                  => gbtBank_txEncodingSel,
          TEST_PATTERN_SEL_I                 => gbtbank_test_pattern_sel_s,
          STATIC_PATTERN_SCEC_I              => "00",
          STATIC_PATTERN_DATA_I              => gbtbank_test_pattern_datawhole_s,
          STATIC_PATTERN_EXTRADATA_WIDEBUS_I => gbtbank_test_pattern_wbuswhole_s,
          -----------------------------------------------
          TX_DATA_O               => txData_from_gbtBank_pattGen(i),
          TX_EXTRA_DATA_WIDEBUS_O => txwBData_from_gbtBank_pattGen(i)
        );

      gbt_txdata_s(i) <= 
        GBTBANK_WB_DATA_I(i-1)(115 downto 32) when 
          -- MG (24.08.2020):
          -- in the original design real data was sent then gbtbank_test_pattern_sel_s was set to 11
          gbtbank_test_pattern_sel_s = "00" and TX_ENCODING = WIDE_BUS else
        GBTBANK_GBT_DATA_I(i-1) when 
          gbtbank_test_pattern_sel_s = "00" and TX_ENCODING = GBT_FRAME else
        txData_from_gbtBank_pattGen(i);

      wb_txdata_s(i) <= GBTBANK_WB_DATA_I(i-1)(31 downto 0) when gbtbank_test_pattern_sel_s = "00" else
        txwBData_from_gbtBank_pattGen(i);

    end generate;
  end generate;

  dataGenDs_gen : if DATA_GENERATOR_ENABLE = DISABLED generate
    dataGenDs_output_gen : for i in 1 to NUM_LINKS generate
      gbt_txdata_s(i) <= GBTBANK_GBT_DATA_I(i-1) when (TX_ENCODING = GBT_FRAME) else GBTBANK_WB_DATA_I(i-1)(115 downto 32);
      wb_txdata_s(i)  <= GBTBANK_WB_DATA_I(i-1)(31 downto 0);

    end generate;
  end generate;

  gbtBank_rxdatamap_gen : for i in 1 to NUM_LINKS generate
    
    GBTBANK_GBT_DATA_O(i-1) <= 
      gbt_rxdata_s(i) when RX_ENCODING = GBT_FRAME else
      (others => '0');

    GBTBANK_WB_DATA_O(i-1) <= 
      gbt_rxdata_s(i) & wb_rxdata_s(i) when RX_ENCODING = WIDE_BUS else
      (others => '0');

  end generate;

  --=============--
  -- Transceiver --
  --=============--   
  gbtBank_mgt_gen : for i in 1 to NUM_LINKS generate

    mgt_devspecific_to_s.drp_addr(i) <= "000000000";
    mgt_devspecific_to_s.drp_en(i)   <= '0';
    mgt_devspecific_to_s.drp_di(i)   <= x"0000";
    mgt_devspecific_to_s.drp_we(i)   <= '0';
    mgt_devspecific_to_s.drp_clk(i)  <= drp_clk;

    mgt_devspecific_to_s.prbs_txSel(i)      <= "000";
    mgt_devspecific_to_s.prbs_rxSel(i)      <= "000";
    mgt_devspecific_to_s.prbs_txForceErr(i) <= '0';
    mgt_devspecific_to_s.prbs_rxCntReset(i) <= '0';

    mgt_devspecific_to_s.conf_diffCtrl(i)   <= "1000";            -- Comment: 807 mVppd
    mgt_devspecific_to_s.conf_postCursor(i) <= "00000";           -- Comment: 0.00 dB (default)
    mgt_devspecific_to_s.conf_preCursor(i)  <= "00000";           -- Comment: 0.00 dB (default)
    mgt_devspecific_to_s.conf_txPol(i)      <= GBTBANK_TX_POL; -- Comment: Not inverted
    mgt_devspecific_to_s.conf_rxPol(i)      <= GBTBANK_RX_POL; -- Comment: Not inverted     

    mgt_devspecific_to_s.loopBack(i) <= GBTBANK_LOOPBACK;

    mgt_devspecific_to_s.rx_p(i) <= GBTBANK_MGT_RX_P(i-1);
    mgt_devspecific_to_s.rx_n(i) <= GBTBANK_MGT_RX_N(i-1);

    mgt_devspecific_to_s.reset_freeRunningClock(i) <= drp_clk;

    GBTBANK_MGT_TX_P(i-1) <= mgt_devspecific_from_s.tx_p(i);
    GBTBANK_MGT_TX_N(i-1) <= mgt_devspecific_from_s.tx_n(i);

    resetOnBitslip_s(i) <= '1' when RX_OPTIMIZATION = LATENCY_OPTIMIZED else '0';
  end generate;

  --============--
  -- GBT Bank   --
  --============--
  gbt_inst : entity work.gbt_bank
    generic map(
      NUM_LINKS       => NUM_LINKS,
      TX_OPTIMIZATION => TX_OPTIMIZATION,
      RX_OPTIMIZATION => RX_OPTIMIZATION,
      TX_ENCODING     => TX_ENCODING,
      RX_ENCODING     => RX_ENCODING
    )
    port map(

      --========--
      -- Resets --
      --========--
      MGT_TXRESET_i => mgt_txreset_s,
      MGT_RXRESET_i => mgt_rxreset_s,
      GBT_TXRESET_i => gbt_txreset_s,
      GBT_RXRESET_i => gbt_rxreset_s,

      --========--
      -- Clocks --     
      --========--
      MGT_CLK_i        => mgt_refclk_i,
      GBT_TXFRAMECLK_i => gbt_txframeclk_s,
      GBT_TXCLKEn_i    => gbt_txclken_s,
      GBT_RXFRAMECLK_i => gbt_rxframeclk_s,
      GBT_RXCLKEn_i    => gbt_rxclken_s,
      MGT_TXWORDCLK_o  => mgt_txwordclk_s,
      MGT_RXWORDCLK_o  => mgt_rxwordclk_s,

      --================--
      -- GBT TX Control --
      --================--
      GBT_ISDATAFLAG_i  => GBTBANK_TX_ISDATA_SEL_I,
      TX_ENCODING_SEL_i => allzeros,

      --=================--
      -- GBT TX Status   --
      --=================--
      TX_PHALIGNED_o  => s_tx_aligned,
      TX_PHCOMPUTED_o => s_tx_aligned_val,

      --================--
      -- GBT RX Control --
      --================--
      RX_ENCODING_SEL_i => allzeros,
      --=================--
      -- GBT RX Status   --
      --=================--
      GBT_RXREADY_o       => gbt_rxready_s,
      --according to gbtxManual header is optional
      GBT_ISDATAFLAG_o    => open,
      GBT_ERRORDETECTED_o => s_gbtbank_rx_errordetected,
      GBT_ERRORFLAG_o     => open,

      --================--
      -- MGT Control    --
      --================--
      MGT_DEVSPECIFIC_i    => mgt_devspecific_to_s,
      MGT_RSTONBITSLIPEn_i => resetOnBitslip_s,
      -- MG: Don't know what it does...
      MGT_RSTONEVEN_i      => mgt_resetoneven,

      --=================--
      -- MGT Status      --
      --=================--
      MGT_TXREADY_o     => mgt_txready_s, --GBTBANK_LINK_TX_READY_O,
      MGT_RXREADY_o     => mgt_rxready_s, --GBTBANK_LINK_RX_READY_O,
      MGT_DEVSPECIFIC_o => mgt_devspecific_from_s,
      MGT_HEADERFLAG_o  => open,
      MGT_RSTCNT_o      => open,
      MGT_HEADERLOCKED_o => mgt_rx_header_locked_s,

      --========--
      -- Data   --
      --========--
      GBT_TXDATA_i => gbt_txdata_s,
      GBT_RXDATA_o => gbt_rxdata_s,

      WB_TXDATA_i => wb_txdata_s,
      WB_RXDATA_o => wb_rxdata_s

    );
    


  
  error_counters_gen : for i in 1 to NUM_LINKS generate
  
    rx_error_counter_inst : entity work.saturated_counter
    generic map (
      g_counter_len => 8
    )
    port map (
      i_clk   => frame_clk_i,
      i_rst   => link_statuses_cleared_stb(i) or gbtbank_general_reset,
      i_pulse => s_gbtbank_rx_errordetected(i),
      std_logic_vector(o_val)   => link_statuses_cleared(i).rx_error_cnt
    );
    
    gbtbank_gbtrx_ready_loss_s(i) <= '1'when gbtbank_gbtrx_ready_s(i)='0' and gbtbank_gbtrx_ready_s(i)='1' else '0';
    
    rx_readyloss_counter_inst : entity work.saturated_counter
    generic map (
      g_counter_len => 8
    )
    port map (
      i_clk   => frame_clk_i,
      i_rst   => link_statuses_cleared_stb(i) or gbtbank_general_reset,
      i_pulse => gbtbank_gbtrx_ready_loss_s(i),
      std_logic_vector(o_val)   => link_statuses_cleared(i).rx_readyloss_cnt
    );

  end generate;

  gbt_rxready_loss_sp : process (frame_clk_i)
  begin
    if (rising_edge(frame_clk_i)) then
      gbtbank_gbtrx_ready_r <= gbtbank_gbtrx_ready_s;
    end if;
  end process gbt_rxready_loss_sp;

  -- place for the registers

  -- 01 - counters on: sc, ec, data and widebus
  -- 10 - static patterns on each field
  -- 11 - all zeros
  -- 00 - actual data
  gbtbank_test_pattern_sel_s      <= pattgen_config.mode;

  -- connect single byte of the pattern to the register
  -- it will be repeated by the GW
  gbtbank_test_pattern_wbusbyte_s <= pattgen_config.const_wbdat_patt;
  gbtbank_test_pattern_databyte_s <= pattgen_config.const_dat_patt;

  gen_wbpatt : for i in 0 to 3 generate
    gbtbank_test_pattern_wbuswhole_s((i+1)*8-1 downto 8*i) <= gbtbank_test_pattern_wbusbyte_s;
  end generate;

  gen_datapat: for i in 0 to 9 generate
    gbtbank_test_pattern_datawhole_s((i+1)*8-1 downto 8*i) <= gbtbank_test_pattern_databyte_s;
  end generate;


  wb_common_reg_inst : entity agwb.gbt_fpga_common
    port map (
      slave_i         => wb_gbtcommon_i,
      slave_o         => wb_gbtcommon_o,
      patgen_o        => pattgen_config,
      ctrl_o.rst(0)      => reset_wb,
      ctrl_o.rst_tx(0)   => gbtbank_manual_reset_tx,
      ctrl_o.rst_rx(0)   => gbtbank_manual_reset_rx,
      ctrl_o.tx_pol(0)   => GBTBANK_TX_POL,
      ctrl_o.rx_pol(0)   => GBTBANK_RX_POL,
      ctrl_o.loopback => GBTBANK_LOOPBACK,
      -- cant use gbtbank_general_reset because it is derived from ctrl_o.rst
      rst_n_i         => not reset_sync,
      clk_sys_i       => frame_clk_i
    );

  link_regs_gen : for i in 1 to NUM_LINKS generate

    wb_link_reg_inst : entity agwb.gbt_fpga_link
    port map (
      slave_i                 => wb_gbtlink_i(i-1),
      slave_o                 => wb_gbtlink_o(i-1),
      status_i                => link_statuses(i),
      status_with_clear_i     => link_statuses_cleared(i),
      status_with_clear_i_ack => link_statuses_cleared_stb(i),
      rst_n_i                 => not gbtbank_general_reset,
      clk_sys_i               => frame_clk_i
    );

    link_statuses(i).rx_header_locked(0) <= mgt_rx_header_locked_s(i);
    link_statuses(i).rx_ready(0)         <= gbtbank_gbtrx_ready_s(i);
    link_statuses(i).tx_phaligned(0)     <= s_tx_aligned(i);
    link_statuses(i).tx_phaligned_val(0) <= s_tx_aligned_val(i);
    link_statuses(i).tx_ready(0)         <= s_tx_ready(i);
    link_statuses(i).link_ready(0)       <= s_link_ready(i);

  end generate;

--=====================================================================================--   
end structural;
--=================================================================================================--
--#################################################################################################--
--=================================================================================================--