library ieee;
    use ieee.std_logic_1164.all;
    use ieee.numeric_std.all;               -- for type conversions

library work;
    use work.gbt_fpga_common_pkg.all;

library general_cores;
    use general_cores.wishbone_pkg.all;


entity multi_crob3_interface is
  generic(
    crobs_g : natural := 8
  );
  port (
    -- experiment reference clock
    -- must be 40 MHz
    -- also used for wishbone
    clk_frame_i          : in  std_logic;

    -- reference clock for the MGT
    -- must be 120 MHz
    -- must come from ibuf (not bufg)
    -- MUST BE SYNCHRONOUS WITH clk_frame_i!!!
    clk_mgtrefs_i        : in  std_logic_vector(crobs_g-1 downto 0);

    --==============--
    -- Reset        --
    --==============--
    -- reset_async_i is synchronized to clk_frame_i inside the core
    -- asserting reset_async_i will set a reset register accessible via wishbone
    -- after toggling reset_async_i user must make sure that clk_frame_i and clk_mgtref_i are stable
    -- and remove reset via gbt_fpga_common ctrl.rst register
    reset_async_i       : in std_logic;

    --==============--
    -- Serial lanes --
    --==============--
    serial_uplinks_i     : in  t_crob3_serial_uplink_arr(crobs_g-1 downto 0);
    serial_downlinks_o   : out t_crob3_serial_downlink_arr(crobs_g-1 downto 0);

    --==============--
    -- Parallel Data--
    --==============--  
    
    -- data to be sent via master downlink (no EC/IC)
    downlinks_i        : in  t_crob3_para_downlink_arr(crobs_g-1 downto 0);

    -- data received via all (master and slave) uplinks (no EC/IC)
    uplinks_o          : out t_crob3_para_uplink_arr(crobs_g-1 downto 0);
    
    --==============--
    -- Wishbone slave interface      --
    --==============--      
    wb_rst_i            : in std_logic;

    -- one WB bus per CROB!
    -- should simplify parametrisation
    wbs_i               : in t_wishbone_slave_in_array(crobs_g-1 downto 0);
    wbs_o               : out t_wishbone_slave_out_array(crobs_g-1 downto 0)

  ) ;
end multi_crob3_interface;

architecture multi_crob3_interface_arc of multi_crob3_interface is

begin

  multiple_crobs_gen : for i in 0 to crobs_g-1 generate
    
    crob_interface_inst : entity work.crob_interface
      generic map (
        links_g => 3
      )
      port map (
        clk_frame_i              => clk_frame_i,
        clk_mgtref_i             => clk_mgtrefs_i(i),
        reset_async_i            => reset_async_i,
        serial_master_downlink_o => serial_downlinks_o(i).master,
        serial_master_uplink_i   => serial_uplinks_i(i).master,
        serial_slave0_downlink_o => serial_downlinks_o(i).slave0,
        serial_slave0_uplink_i   => serial_uplinks_i(i).slave0,
        serial_slave1_downlink_o => serial_downlinks_o(i).slave1,
        serial_slave1_uplink_i   => serial_uplinks_i(i).slave1,
        downlink_i               => downlinks_i(i),
        uplink_master_o          => uplinks_o(i).master,
        uplink_slave0_o          => uplinks_o(i).slave0,
        uplink_slave1_o          => uplinks_o(i).slave1,
        wb_rst_i                 => wb_rst_i,
        wbs_i                    => wbs_i(i),
        wbs_o                    => wbs_o(i)
      );  
  end generate;

end  multi_crob3_interface_arc;